/**
 * \file tagger.h
 * Class for reading, writing and storing tags.
 *
 * \b Project: Qoobar
 * \author Alex Novichkov
 * \date 8 Dec 2009
 *
 * Copyright (C) 2009  Alex Novichkov
 *
 * This file is part of Qoobar.
 *
 * Qoobar is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Qoobar is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef TAGGER_H
#define TAGGER_H

#include <QString>
#include <QMap>
#include <QSharedData>
#include <QMetaType>
#include "enums.h"
#include "coverimage.h"

void append(QString &tagString, const QString &newTag, const QString &sep=QSL(";"));

struct ReplayGainInfo {
    ReplayGainInfo() {}
    ReplayGainInfo(const ReplayGainInfo &rg) :
        albumPeak(rg.albumPeak),
        trackPeak(rg.trackPeak),
        albumGain(rg.albumGain),
        trackGain(rg.trackGain),
        albumMinMax(rg.albumMinMax),
        trackMinMax(rg.trackMinMax),
        loudness(rg.loudness),
        undo(rg.undo)
    {}
    void clear();
    bool isEmpty() const;
    bool operator ==(const ReplayGainInfo &rg) const {
        return (albumPeak == rg.albumPeak
                && albumGain == rg.albumGain
                && trackPeak == rg.trackPeak
                && trackGain == rg.trackGain
                && albumMinMax == rg.albumMinMax
                && trackMinMax == rg.trackMinMax
                && loudness == rg.loudness
                && undo == rg.undo);
    }
    bool operator !=(const ReplayGainInfo &rg) const {
        return (albumPeak!=rg.albumPeak
                || albumGain!=rg.albumGain
                || trackPeak!=rg.trackPeak
                || trackGain!=rg.trackGain
                || albumMinMax!=rg.albumMinMax
                || trackMinMax!=rg.trackMinMax
                || loudness!=rg.loudness
                || undo!=rg.undo);
    }

    QString albumPeak;
    QString trackPeak;
    QString albumGain;
    QString trackGain;
    QString albumMinMax;
    QString trackMinMax;
    QString loudness;
    QString undo;
};

class TagData : public QSharedData
{
public:
    TagData();
    TagData(const TagData &other);
    ~TagData() {}

    QString icon;
    QString fullFileName; //full path to file
    QString fileNameExt; //file name with extension
    QString fileName; //file name without extension
    QString filePath;
    QString fileExt; //extension without dot
    QString bitrate;
    int length; //length in seconds
    int sampleRate;
    int channels;
    bool wasChanged;
    int filetype;
    bool readOnly;

    QString album; //a
    QString artist; //A
    QString title; //t
    QString performer; //p
    QString comment; //C
    QString genre; //g
    QString composer; //c
    QString year; //y
    QString tracknumber; //n
    QString totalTracks; //T
    QString conductor; //d
    QString orchestra; //o
    QString subtitle; //s
    QString key; //k
    QString albumArtist; //u

    QString category; //G
    QString publisher; //P
    QString copyright; //X
    QString mood; //m
    QString tempo; //B
    QString lyricist; //w
    QString lyrics; //W
    QString discnumber; //D
    QString totaldiscs; //S
    QString encodedby; //e
    QString remixedby; //x
    QString rating; //R
    QString originalalbum; //O
    QString originalartist; //v
    QString originallyricist; //i
    QString url; //U
    QString isrc; //I

    ReplayGainInfo replayGainInfo;
    CoverImage image;
    QMap<QString,QString> otherTags;
};

class Tag
{
public:
    enum FileType {
        MP3_FILE=0,
        OGG_FILE=1,
        WMA_FILE=2,
        FLAC_FILE=3,
        M4A_FILE=4,
        MPC_FILE=5,
        WV_FILE=6,
        WAV_FILE=7,
        SPX_FILE=8,
        TTA_FILE=9,
        OGA_FILE=10,
        APE_FILE=11,
        AIFF_FILE=12,
        OPUS_FILE=13,
        DSF_FILE=14,
        FILES_NUM=15
    };
    Tag();
    Tag(const QString &filename);
    Tag(const Tag &other) : d(other.d) {}

    void clear();

    ReplayGainInfo replayGainInfo() const {return d->replayGainInfo;}
    void setReplayGainInfo(const ReplayGainInfo &rg, bool updateStatus=true);
    void removeReplayGainInfo(bool updateStatus=true);
    bool replayGainInfoIsEmpty() const {return d->replayGainInfo.isEmpty();}
    void setAlbumPeak(const QString &value, bool updateStatus=true);
    void setAlbumGain(const QString &value, bool updateStatus=true);
    void setTrackPeak(const QString &value, bool updateStatus=true);
    void setTrackGain(const QString &value, bool updateStatus=true);
    void setAlbumMinMax(const QString &value, bool updateStatus=true);
    void setTrackMinMax(const QString &value, bool updateStatus=true);
    void setLoudness(const QString &value, bool updateStatus=true);
    void setReplayGainUndo(const QString &value, bool updateStatus=true);

    CoverImage image() const;
    void setImage(const CoverImage &image, bool updateStatus=true);
    void removeImage(bool updateStatus=true);
    bool imageIsEmpty() const {return d->image.isEmpty();}
    void setImagePixmap(const QByteArray &pixmap, bool updateStatus=true);
    void setImageDescription(const QString &description, bool updateStatus=true);
    void setImageMimetype(const QString &mimetype, bool updateStatus=true);
    void setImageType(int type, bool updateStatus=true);


    void setFile(const QString & fullFileName);
    /**
     * @brief fullFileName
     * @return full path of file
     */
    QString fullFileName() const {return d->fullFileName;}

    int fileType() const {return d->filetype;}
    /**
     * @brief fileNameExt
     * @return  file name with extension
     */
    QString fileNameExt() const {return d->fileNameExt;}
    /**
     * @brief filePath
     * @return file path without trailing slash
     */
    QString filePath() const {return d->filePath;}
    /**
     * @brief fileExt
     * @return  file name extension without leading dot
     */
    QString fileExt() const {return d->fileExt;}
    /**
     * @brief fileName
     * @return file name without extension
     */
    QString fileName() const {return d->fileName;}


    int length() const {return d->length;}
    void setLength(int length);

    QString bitrate() const {return d->bitrate;}
    void setBitrate(const QString &bitrate) {d->bitrate=bitrate;}

    int sampleRate() const {return d->sampleRate;}
    void setSampleRate(int sampleRate) {d->sampleRate = sampleRate;}

    int channels() const {return d->channels;}
    void setChannels(int channels) {d->channels = channels;}

    QString icon() const {return d->icon;}

    bool wasChanged() const {return d->wasChanged;}
    void setChanged(const bool changed);

    bool readOnly() const {return d->readOnly;}


    QString tag(const int &) const;
    QString nonEmptyTag(const int & tagID,const QString &defaultVal) const;
    QString userTag(const QString &) const;
    QStringList userTagsKeys() const;
    QMap<QString,QString> userTags() const {return d->otherTags;}

    void setTag(const int tagID,const QString &newTag);
    void setTag1(const int tagID,const QString &newTag);
    void appendTag(const int tagID,const QString &newTag);
    void setUserTag(const QString &,const QString &);
    void appendUserTag(const QString &,const QString &);
    void setUserTags1(const QMap<QString,QString> &other);

    void removeStandardTags(bool updateStatus=true);
    void removeAllTags(bool updateStatus=true);
    void removeUserTags(bool updateStatus=true);

    QString album() const {return d->album;} //a
    QString artist() const {return d->artist;} //A
    QString title() const {return d->title;} //t
    QString performer() const {return d->performer;} //p
    QString comment() const {return d->comment;} //C
    QString genre() const {return d->genre;} //g
    QString composer() const {return d->composer;} //c
    QString year() const {return d->year;} //y
    QString tracknumber() const {return d->tracknumber;} //n
    QString totalTracks() const {return d->totalTracks;} //T
    QString conductor() const {return d->conductor;} //d
    QString orchestra() const {return d->orchestra;} //o
    QString subtitle() const {return d->subtitle;} //s
    QString key() const {return d->key;} //k
    QString albumArtist() const {return d->albumArtist;} //u
    QString category() const {return d->category;} //G
    QString publisher() const {return d->publisher;} //P
    QString copyright() const {return d->copyright;} //X
    QString mood() const {return d->mood;} //m
    QString tempo() const {return d->tempo;} //B
    QString lyricist() const {return d->lyricist;} //w
    QString lyrics() const {return d->lyrics;} //W
    QString discnumber() const {return d->discnumber;} //D
    QString totaldiscs() const {return d->totaldiscs;} //S
    QString encodedby() const {return d->encodedby;} //e
    QString remixedby() const {return d->remixedby;} //x
    QString rating() const {return d->rating;} //R
    QString originalalbum() const {return d->originalalbum;} //O
    QString originalartist() const {return d->originalartist;} //v
    QString originallyricist() const {return d->originallyricist;} //i
    QString url() const {return d->url;} //U
    QString isrc() const {return d->isrc;} //I

    bool operator==(const Tag &t) const;
    bool operator!=(const Tag &t) const;
private:
    void setTag(QString &tag,const QString &newTag);
    QSharedDataPointer<TagData> d;
};

Q_DECLARE_METATYPE(Tag)


#endif
